/*
* Copyright 1993-2007 NVIDIA Corporation.  All rights reserved.
*
* NOTICE TO USER:
*
* This source code is subject to NVIDIA ownership rights under U.S. and
* international Copyright laws.
*
* NVIDIA MAKES NO REPRESENTATION ABOUT THE SUITABILITY OF THIS SOURCE
* CODE FOR ANY PURPOSE.  IT IS PROVIDED "AS IS" WITHOUT EXPRESS OR
* IMPLIED WARRANTY OF ANY KIND.  NVIDIA DISCLAIMS ALL WARRANTIES WITH
* REGARD TO THIS SOURCE CODE, INCLUDING ALL IMPLIED WARRANTIES OF
* MERCHANTABILITY, NONINFRINGEMENT, AND FITNESS FOR A PARTICULAR PURPOSE.
* IN NO EVENT SHALL NVIDIA BE LIABLE FOR ANY SPECIAL, INDIRECT, INCIDENTAL,
* OR CONSEQUENTIAL DAMAGES, OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS
* OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE
* OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE
* OR PERFORMANCE OF THIS SOURCE CODE.
*
* U.S. Government End Users.  This source code is a "commercial item" as
* that term is defined at 48 C.F.R. 2.101 (OCT 1995), consisting  of
* "commercial computer software" and "commercial computer software
* documentation" as such terms are used in 48 C.F.R. 12.212 (SEPT 1995)
* and is provided to the U.S. Government only as a commercial end item.
* Consistent with 48 C.F.R.12.212 and 48 C.F.R. 227.7202-1 through
* 227.7202-4 (JUNE 1995), all U.S. Government End Users acquire the
* source code with only those rights set forth herein.
*/

/*
    Image box filtering example

    This sample uses CUDA to perform a simple box filter on an image
    and uses OpenGL to display the results.

    It processes rows and columns of the image in parallel.

    The box filter is implemented such that it has a constant cost,
    regardless of the filter width.

    Press '=' to increment the filter radius, '-' to decrease it
*/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <math.h>
#include <GL/glew.h>
#include <GL/glut.h>
#include <cuda_gl_interop.h>
#include <cutil.h>

#include <boxFilter_kernel.cu>

char *image_filename = "lena_bw.pgm";
int iterations = 1;
int filter_radius = 10;
int nthreads = 32;

unsigned int width, height;
float* h_img = NULL;
float* d_img = NULL;
float* d_temp = NULL;
cudaArray* d_array, *d_tempArray;

GLuint pbo;     // OpenGL pixel buffer object
GLuint texid;   // texture
GLuint shader;

unsigned int timer = 0;
int fpsCount = 0;        // FPS count for averaging
int fpsLimit = 1;        // FPS limit for sampling

extern "C"
void computeGold(float *id, float *od, int w, int h, int n);

/*
    Perform 2D box filter on image using CUDA

    Parameters:
    d_src  - pointer to input image in device memory
    d_temp - pointer to temporary storage in device memory
    d_dest - pointer to destination image in device memory
    width  - image width
    height - image height
    radius - filter radius
    iterations - number of iterations

*/
void boxFilter(float *d_src, float *d_temp, float *d_dest, int width, int height, int radius, int iterations)
{
    CUDA_SAFE_CALL( cudaBindTextureToArray(tex, d_array) );

    for(int i=0; i<iterations; i++) {
        // use texture for horizontal pass
        d_boxfilter_x_tex<<< height / nthreads, nthreads, 0 >>>( d_temp, width, height, radius);
        d_boxfilter_y_global<<< width / nthreads, nthreads, 0 >>>( d_temp, d_dest, width, height, radius);
        if (iterations > 1) {
            // copy result back from global memory to array
            CUDA_SAFE_CALL( cudaMemcpyToArray( d_tempArray, 0, 0, d_dest, width * height * sizeof(float), cudaMemcpyDeviceToDevice));
            CUDA_SAFE_CALL( cudaBindTextureToArray(tex, d_tempArray) );
        }
    }
}

// display results using OpenGL
void display()
{
    CUT_SAFE_CALL(cutStartTimer(timer));  

    // execute filter, writing results to pbo
    float *d_result;
    CUDA_SAFE_CALL(cudaGLMapBufferObject((void**)&d_result, pbo));
    boxFilter(d_img, d_temp, d_result, width, height, filter_radius, iterations);
    CUDA_SAFE_CALL(cudaGLUnmapBufferObject(pbo));

    // display results
    glClear(GL_COLOR_BUFFER_BIT);

    // load texture from pbo
    glBindBufferARB(GL_PIXEL_UNPACK_BUFFER_ARB, pbo);
    glBindTexture(GL_TEXTURE_2D, texid);
    glTexSubImage2D(GL_TEXTURE_2D, 0, 0, 0, width, height, GL_LUMINANCE, GL_FLOAT, 0);
    glBindBufferARB(GL_PIXEL_UNPACK_BUFFER_ARB, 0);

    // fragment program is required to display floating point texture
    glBindProgramARB(GL_FRAGMENT_PROGRAM_ARB, shader);
    glEnable(GL_FRAGMENT_PROGRAM_ARB);
    glDisable(GL_DEPTH_TEST);

    glBegin(GL_QUADS);
    glVertex2f(0, 0); glTexCoord2f(0, 0);
    glVertex2f(0, 1); glTexCoord2f(1, 0);
    glVertex2f(1, 1); glTexCoord2f(1, 1);
    glVertex2f(1, 0); glTexCoord2f(0, 1);
    glEnd();
    glBindTexture(GL_TEXTURE_2D, 0);
    glDisable(GL_FRAGMENT_PROGRAM_ARB);

    glutSwapBuffers();

    CUT_SAFE_CALL(cutStopTimer(timer));  

    fpsCount++;
    if (fpsCount == fpsLimit) {
        char fps[256];
        float ifps = 1.f / (cutGetAverageTimerValue(timer) / 1000.f);
        sprintf(fps, "CUDA box filter: %3.1f fps", ifps);
        glutSetWindowTitle(fps);
        fpsCount = 0; 
        fpsLimit = (int)max(ifps, 1.f);
        CUT_SAFE_CALL(cutResetTimer(timer));  
    }
}

void idle()
{
    glutPostRedisplay();
}

void keyboard(unsigned char key, int x, int y)
{
    switch(key) {
        case 27:
            exit(0);
            break;
        case '=':
        case '+':
            filter_radius++;
            break;
        case '-':
            if (filter_radius > 1) filter_radius--;
            break;
        case ']':
            iterations++;
            break;
        case '[':
            if (iterations>1) iterations--;
            break;
        default:
            break;
    }
    printf("radius = %d, iterations = %d\n", filter_radius, iterations);
    glutPostRedisplay();
}

void reshape(int x, int y)
{
    glViewport(0, 0, x, y);

    glMatrixMode(GL_MODELVIEW);
    glLoadIdentity();

    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();
    glOrtho(0.0, 1.0, 0.0, 1.0, 0.0, 1.0); 
}

void initCuda()
{
    unsigned int size = width * height * sizeof(float);

    // allocate device memory
    CUDA_SAFE_CALL( cudaMalloc( (void**) &d_img, size));
    CUDA_SAFE_CALL( cudaMalloc( (void**) &d_temp, size));

    // copy image data to array
    cudaChannelFormatDesc channelDesc = cudaCreateChannelDesc(32, 0, 0, 0, cudaChannelFormatKindFloat);
    CUDA_SAFE_CALL( cudaMallocArray( &d_array, &channelDesc, width, height )); 
    CUDA_SAFE_CALL( cudaMemcpyToArray( d_array, 0, 0, h_img, size, cudaMemcpyHostToDevice));

    CUDA_SAFE_CALL( cudaMallocArray( &d_tempArray, &channelDesc, width, height )); 

    // set texture parameters
    tex.addressMode[0] = cudaAddressModeClamp;
    tex.addressMode[1] = cudaAddressModeClamp;
    tex.filterMode = cudaFilterModePoint;
    tex.normalized = false;

    // Bind the array to the texture
    CUDA_SAFE_CALL( cudaBindTextureToArray(tex, d_array, channelDesc) );

    CUT_SAFE_CALL( cutCreateTimer( &timer));
}

void cleanup()
{
    CUT_SAFE_CALL( cutDeleteTimer( timer));
    free(h_img);

    CUDA_SAFE_CALL(cudaFree(d_img));
    CUDA_SAFE_CALL(cudaFree(d_temp));
    CUDA_SAFE_CALL(cudaFreeArray(d_array));
    CUDA_SAFE_CALL(cudaFreeArray(d_tempArray));

	CUDA_SAFE_CALL(cudaGLUnregisterBufferObject(pbo));    
	glDeleteBuffersARB(1, &pbo);
    glDeleteTextures(1, &texid);
    glDeleteProgramsARB(1, &shader);
}

// shader for displaying floating-point texture
static const char *shader_code = 
"!!ARBfp1.0\n"
"TEX result.color, fragment.texcoord, texture[0], 2D; \n"
"END";

GLuint compileASMShader(GLenum program_type, const char *code)
{
    GLuint program_id;
    glGenProgramsARB(1, &program_id);
    glBindProgramARB(program_type, program_id);
    glProgramStringARB(program_type, GL_PROGRAM_FORMAT_ASCII_ARB, (GLsizei) strlen(code), (GLubyte *) code);

    GLint error_pos;
    glGetIntegerv(GL_PROGRAM_ERROR_POSITION_ARB, &error_pos);
    if (error_pos != -1) {
        const GLubyte *error_string;
        error_string = glGetString(GL_PROGRAM_ERROR_STRING_ARB);
        fprintf(stderr, "Program error at position: %d\n%s\n", (int)error_pos, error_string);
        return 0;
    }
    return program_id;
}

void initOpenGL()
{
    // create pixel buffer object
    glGenBuffersARB(1, &pbo);
	glBindBufferARB(GL_PIXEL_UNPACK_BUFFER_ARB, pbo);
	glBufferDataARB(GL_PIXEL_UNPACK_BUFFER_ARB, width*height*sizeof(float), h_img, GL_STREAM_DRAW_ARB);

	glBindBufferARB(GL_PIXEL_UNPACK_BUFFER_ARB, 0);
	CUDA_SAFE_CALL(cudaGLRegisterBufferObject(pbo));

    // create texture for display
    glGenTextures(1, &texid);
    glBindTexture(GL_TEXTURE_2D, texid);
    glTexImage2D(GL_TEXTURE_2D, 0, GL_LUMINANCE32F_ARB, width, height, 0, GL_LUMINANCE, GL_FLOAT, NULL);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
    glBindTexture(GL_TEXTURE_2D, 0);

    // load shader program
    shader = compileASMShader(GL_FRAGMENT_PROGRAM_ARB, shader_code);
}

////////////////////////////////////////////////////////////////////////////////
//! Run a simple test for CUDA
////////////////////////////////////////////////////////////////////////////////
void
runTest( int argc, char** argv) 
{
    // warm-up
    boxFilter(d_img, d_temp, d_temp, width, height, filter_radius, iterations);

    CUDA_SAFE_CALL( cudaThreadSynchronize() );
    CUT_SAFE_CALL( cutStartTimer( timer));

    // execute the kernel
    boxFilter(d_img, d_temp, d_img, width, height, filter_radius, iterations);

    // check if kernel execution generated an error
    CUT_CHECK_ERROR("Kernel execution failed");

    CUDA_SAFE_CALL( cudaThreadSynchronize() );
    CUT_SAFE_CALL( cutStopTimer( timer));

    printf("Processing time: %f (ms)\n", cutGetTimerValue( timer));
	printf("%.2f Mpixels/sec\n", (width*height / (cutGetTimerValue( timer) / 1000.0f)) / 1e6);

    // allocate mem for the result on host side
    unsigned int size = width * height * sizeof(float);
    float* h_odata = (float*) malloc(size);
    // copy result from device to host
    CUDA_SAFE_CALL( cudaMemcpy( h_odata, d_img, size, cudaMemcpyDeviceToHost) );

    CUT_SAFE_CALL( cutSavePGMf( "data/output.pgm", h_odata, width, height));

    // compute reference solution
    float *h_odata_ref = (float*) malloc(size);
    CUT_SAFE_CALL( cutStartTimer( timer));
    computeGold(h_img, h_odata_ref, width, height, filter_radius);
    CUT_SAFE_CALL( cutStopTimer( timer));
    printf("CPU Processing time: %f (ms)\n", cutGetTimerValue( timer));

    // check result
    if( cutCheckCmdLineFlag( argc, (const char**) argv, "regression")) 
    {
        // write file for regression test
        CUT_SAFE_CALL( cutSavePGMf( "data/regression.pgm", h_odata, width, height));
        printf("Wrote 'regression.pgm'\n");
    }
    else 
    {
        // custom output handling when no regression test running
        // in this case check if the result is equivalent to the expected soluion
        CUTBoolean res = cutComparef( h_odata, h_odata_ref, width*height);
        printf( "Test %s\n", (1 == res) ? "PASSED" : "FAILED");
    }

    free(h_odata);
    free(h_odata_ref);
}

////////////////////////////////////////////////////////////////////////////////
// Program main
////////////////////////////////////////////////////////////////////////////////
int
main( int argc, char** argv) 
{

    CUT_DEVICE_INIT();

    cutGetCmdLineArgumenti( argc, (const char**) argv, "threads", &nthreads);
    cutGetCmdLineArgumenti( argc, (const char**) argv, "radius", &filter_radius);

    // load image
    char* image_path = cutFindFilePath(image_filename, argv[0]);
    if (image_path == 0) {
        fprintf(stderr, "Error finding image file '%s'\n", image_filename);
        exit(EXIT_FAILURE);
    }

    CUT_SAFE_CALL( cutLoadPGMf(image_path, &h_img, &width, &height));
    if (!h_img) {
        printf("Error opening file '%s'\n", image_path);
        exit(-1);
    }
    printf("Loaded '%s', %d x %d pixels\n", image_path, width, height);

    // initialize GLUT
    glutInit(&argc, argv);
    glutInitDisplayMode(GLUT_RGB | GLUT_DOUBLE);
    glutInitWindowSize(width, height);
    glutCreateWindow("CUDA box filter");
    glutDisplayFunc(display);
    glutKeyboardFunc(keyboard);
    glutReshapeFunc(reshape);
    glutIdleFunc(idle);

    printf("Press '+' and '-' to change filter width\n"
           "Press ']' and '[' to change number of iterations\n");

    glewInit();
    if (!glewIsSupported("GL_VERSION_2_0 GL_VERSION_1_5 GL_ARB_vertex_buffer_object GL_ARB_pixel_buffer_object")) {
        fprintf(stderr, "Required OpenGL extensions missing.");
        exit(-1);
    }
    initCuda();
    initOpenGL();

    atexit(cleanup);

//    runTest( argc, argv);
    glutMainLoop();
    return 0;
}
